from collections.abc import Sequence
from typing import Any

from .definitions import COLOR_NAME_TO_RGB, LONG_HEX_COLOR, SHORT_HEX_COLOR


def is_long_hex(color: str) -> bool:
    return bool(LONG_HEX_COLOR.fullmatch(color))


def is_short_hex(color: str) -> bool:
    return bool(SHORT_HEX_COLOR.fullmatch(color))


def is_rgb(color: Any) -> bool:
    if not isinstance(color, Sequence) or isinstance(color, str):
        return False
    if len(color) != 3:
        return False
    for channel in color:
        if not isinstance(channel, int | float) or not (0 <= channel <= 255):
            return False
    return True


def is_rgbf(color: Any) -> bool:
    if not isinstance(color, Sequence) or isinstance(color, str):
        return False
    if len(color) != 3:
        return False
    for channel in color:
        if not isinstance(channel, int | float) or not (0 <= channel <= 1):
            return False
    return True


def is_rgba(color: Any) -> bool:
    if not isinstance(color, Sequence) or isinstance(color, str):
        return False
    if len(color) != 4:
        return False
    for channel in color:
        if not isinstance(channel, int | float) or not (0 <= channel <= 255):
            return False
    return True


def is_rgbaf(color: Any) -> bool:
    if not isinstance(color, Sequence) or isinstance(color, str):
        return False
    if len(color) != 4:
        return False
    for channel in color:
        if not isinstance(channel, int | float) or not (0 <= channel <= 1):
            return False
    return True


def is_web(color: str) -> bool:
    return color in COLOR_NAME_TO_RGB or is_long_hex(color) or is_short_hex(color)


def is_hsl(color: Any) -> bool:
    if not isinstance(color, Sequence) or isinstance(color, str):
        return False
    if len(color) != 3:
        return False
    if isinstance(color[0], int | float) and not 0 <= color[0] <= 360:
        return False
    for channel in color[1:]:
        if not isinstance(channel, int | float) or not (0 <= channel <= 1):
            return False
    return True


def is_hsla(color: Any) -> bool:
    if not isinstance(color, Sequence) or isinstance(color, str):
        return False
    if len(color) != 4:
        return False
    for channel in color:
        if not isinstance(channel, int | float) or not (0 <= channel <= 1):
            return False
    return True



from colourings.identify import is_hsl, is_hsla, is_rgb, is_rgba, is_rgbaf, is_rgbf


def test_bad_rbg():
    assert not is_rgb((300, 0, 0))
    assert not is_rgb((30, 300, 0, 0))
    assert not is_rgb("30, 300, 0, 0")
    assert not is_rgb(int)


def test_bad_rbgf():
    assert not is_rgbf((1.1, 0, 0))
    assert not is_rgbf((1.1, 2, 0, 0))
    assert not is_rgbf("30, 300, 0, 0")
    assert not is_rgbf(int)


def test_bad_rbgaf():
    assert not is_rgbaf((1.1, 0, 0))
    assert not is_rgbaf((1.1, 2, 0, 0))
    assert not is_rgbaf("30, 300, 0, 0")
    assert not is_rgbaf(int)


def test_bad_rbga():
    assert not is_rgba((255.1, 0, 0))
    assert not is_rgba((255.1, 2, 0, 0))
    assert not is_rgba("30, 300, 0, 0")
    assert not is_rgba(int)


def test_bad_hsl():
    assert not is_hsl((400, 0, 0))
    assert not is_hsl((30, 300, 0, 0))
    assert not is_hsl("30, 300, 0, 0")
    assert not is_hsl(int)


def test_bad_hsla():
    assert not is_hsla((1.1, 0, 0))
    assert not is_hsla((1.1, 2, 0, 0))
    assert not is_hsla("30, 300, 0, 0")
    assert not is_hsla(int)



import re
from collections.abc import Sequence

from .definitions import (
    COLOR_NAME_TO_RGB,
    FLOAT_ERROR,
    LONG_HEX_COLOR,
    RGB_TO_COLOR_NAMES,
    SHORT_HEX_COLOR,
)
from .identify import (
    is_hsl,
    is_hsla,
    is_long_hex,
    is_rgb,
    is_rgba,
    is_rgbaf,
    is_rgbf,
    is_short_hex,
    is_web,
)

# add HSV, CMYK, YUV conversion


def rgb2rgba(
    rgb: Sequence[int | float], alpha: int | float
) -> tuple[float, float, float, float]:
    return rgb[0], rgb[1], rgb[2], alpha * 255.0


def rgb2rgbf(rgb: Sequence[int | float]) -> tuple[float, float, float]:
    return rgb[0] / 255.0, rgb[1] / 255.0, rgb[2] / 255.0


def rgb2rgbaf(
    rgb: Sequence[int | float], alpha: int | float
) -> tuple[float, float, float, float]:
    return rgb[0] / 255.0, rgb[1] / 255.0, rgb[2] / 255.0, alpha


def hsl2hsla(
    hsl: Sequence[int | float], alpha: int | float
) -> tuple[float, float, float, float]:
    return hsl[0] / 360.0, hsl[1], hsl[2], alpha


def hsl2rgb(hsl: Sequence[int | float]) -> tuple[float, float, float]:
    """Convert HSL representation towards RGB

    :param h: Hue, position around the chromatic circle (h=1 equiv h=0)
    :param s: Saturation, color saturation (0=full gray, 1=full color)
    :param l: Ligthness, Overhaul lightness (0=full black, 1=full white)
    :rtype: 3-uple for RGB values in float between 0 and 1

    Hue, Saturation, Range from Lightness is a float between 0 and 1

    Note that Hue can be set to any value but as it is a rotation
    around the chromatic circle, any value above 1 or below 0 can
    be expressed by a value between 0 and 1 (Note that h=0 is equiv
    to h=1).

    This algorithm came from:
    http://www.easyrgb.com/index.php?X=MATH&H=19#text19
    """
    if not is_hsl(hsl):
        raise ValueError("Input is not an HSL type.")

    _h, _s, _l = [float(v) for v in hsl]
    _h /= 360.0

    if _s == 0:
        return _l * 255.0, _l * 255.0, _l * 255.0

    v2 = _l * (1.0 + _s) if _l < 0.5 else (_l + _s) - (_s * _l)

    v1 = 2.0 * _l - v2

    r = _hue2rgb(v1, v2, _h + (1.0 / 3))
    g = _hue2rgb(v1, v2, _h)
    b = _hue2rgb(v1, v2, _h - (1.0 / 3))

    return r * 255.0, g * 255.0, b * 255.0


def hsl2rgbf(hsl: Sequence[int | float]) -> tuple[float, float, float]:
    return rgb2rgbf(hsl2rgb(hsl))


def rgba2hsl(rgba: Sequence[int | float]) -> tuple[float, float, float]:
    if not is_rgba(rgba):
        raise ValueError("Input is not an RGBA type.")
    return rgb2hsl(rgba[:3])


def rgbaf2hsl(rgbaf: Sequence[int | float]) -> tuple[float, float, float]:
    if not is_rgbaf(rgbaf):
        raise ValueError("Input is not an RGBAf type.")
    rgb = [v * 255.0 for v in rgbaf[:3]]
    return rgb2hsl(rgb)


def hsla2hsl(hsla: Sequence[int | float]) -> tuple[float, float, float]:
    if not is_hsla(hsla):
        raise ValueError("Input is not an HSLA type.")
    hsl = (hsla[0] * 360, hsla[1], hsla[2])
    return hsl


def rgbf2hsl(rgbf: Sequence[int | float]) -> tuple[float, float, float]:
    if not is_rgbf(rgbf):
        raise ValueError("Input is not an RGBf type.")
    rgb = [v * 255.0 for v in rgbf]
    return rgb2hsl(rgb)


def rgb2hsl(rgb: Sequence[int | float]) -> tuple[float, float, float]:
    """Convert RGB representation towards HSL

    :param r: Red amount (float between 0 and 255)
    :param g: Green amount (float between 0 and 255)
    :param b: Blue amount (float between 0 and 255)
    :rtype: 3-uple for HSL values in float between 0 and 255

    This algorithm came from:
    http://www.easyrgb.com/index.php?X=MATH&H=19#text19
    """
    if not is_rgb(rgb):
        raise ValueError("Input is not an RGB type.")
    r, g, b = [float(v) / 255 for v in rgb]

    vmin = min(r, g, b)  ## Min. value of RGB
    vmax = max(r, g, b)  ## Max. value of RGB
    diff = vmax - vmin  ## Delta RGB value

    vsum = vmin + vmax

    _l = vsum / 2

    if diff < FLOAT_ERROR:  ## This is a gray, no chroma...
        return (0.0, 0.0, _l)

    ##
    ## Chromatic data...
    ##

    ## Saturation
    s = diff / vsum if _l < 0.5 else diff / (2.0 - vsum)

    dr = (((vmax - r) / 6) + (diff / 2)) / diff
    dg = (((vmax - g) / 6) + (diff / 2)) / diff
    db = (((vmax - b) / 6) + (diff / 2)) / diff

    if r == vmax:
        h = db - dg
    elif g == vmax:
        h = (1.0 / 3) + dr - db
    elif b == vmax:
        h = (2.0 / 3) + dg - dr

    if h < 0:
        h += 1
    if h > 1:
        h -= 1

    return h * 360.0, s, _l


def _hue2rgb(v1, v2, vH):
    """Private helper function (Do not call directly)

    :param vH: rotation around the chromatic circle (between 0..1)

    """

    while vH < 0:
        vH += 1
    while vH > 1:
        vH -= 1

    if 6 * vH < 1:
        return v1 + (v2 - v1) * 6 * vH
    if 2 * vH < 1:
        return v2
    if 3 * vH < 2:
        return v1 + (v2 - v1) * ((2.0 / 3) - vH) * 6

    return v1


def rgb2hex(rgb: Sequence[int | float], force_long: bool = False) -> str:
    """Transform RGB tuple to hex RGB representation

    :param rgb: RGB 3-uple of float between 0 and 1
    :rtype: 3 hex char or 6 hex char string representation
    """
    if not is_rgb(rgb):
        raise ValueError("Input is not of RGB type.")

    hx = "".join([f"{int(c + 0.5 - FLOAT_ERROR):02x}" for c in rgb])

    if not force_long and hx[0::2] == hx[1::2]:
        hx = "".join(hx[0::2])

    return f"#{hx}"


def hex2rgb(hex: str) -> tuple[float, float, float]:
    """Transform hex RGB representation to RGB tuple

    :param str_rgb: 3 hex char or 6 hex char string representation
    :rtype: RGB 3-uple of float between 0 and 1
    """

    if not (is_long_hex(hex) or is_short_hex(hex)):
        raise ValueError("Input is not of hex type.")

    try:
        rgb = hex[1:]

        if len(rgb) == 6:
            r, g, b = rgb[0:2], rgb[2:4], rgb[4:6]
        elif len(rgb) == 3:
            r, g, b = rgb[0] * 2, rgb[1] * 2, rgb[2] * 2
        else:
            raise ValueError("Length of rgb must be either three or six.")
    except Exception as e:
        raise ValueError(f"Invalid value {hex} provided for rgb color.") from e

    return (float(int(r, 16)), float(int(g, 16)), float(int(b, 16)))


def hex2web(hex: str) -> str:
    """Converts HEX representation to WEB

    :param rgb: 3 hex char or 6 hex char string representation
    :rtype: web string representation (human readable if possible)

    WEB representation uses X11 rgb.txt to define conversion
    between RGB and english color names.
    """
    if not (is_long_hex(hex) or is_short_hex(hex)):
        raise ValueError("Input is not of hex type.")

    rgb = hex2rgb(hex)
    dec_rgb = (int(rgb[0]), int(rgb[1]), int(rgb[2]))
    if dec_rgb in RGB_TO_COLOR_NAMES:
        ## take the first one
        color_name = RGB_TO_COLOR_NAMES[dec_rgb][0]
        ## Enforce full lowercase for single worded color name.
        return (
            color_name
            if len(re.sub(r"[^A-Z]", "", color_name)) > 1
            else color_name.lower()
        )

    # Hex format is verified by hex2rgb function. And should be 3 or 6 digit
    if len(hex) == 7 and hex[1] == hex[2] and hex[3] == hex[4] and hex[5] == hex[6]:
        return "#" + hex[1] + hex[3] + hex[5]
    return hex


def web2hex(web: str, force_long=False) -> str:
    """Converts WEB representation to HEX

    :param rgb: web string representation (human readable if possible)
    :rtype: 3 hex char or 6 hex char string representation

    WEB representation uses X11 rgb.txt to define conversion
    between RGB and english color names.
    """
    web = web.lower()
    if web.startswith("#"):
        if LONG_HEX_COLOR.match(web) or (not force_long and SHORT_HEX_COLOR.match(web)):
            return web.lower()
        elif SHORT_HEX_COLOR.match(web) and force_long:
            return "#" + "".join([str(t) * 2 for t in web[1:]])
        raise AttributeError(f"{web} is not in web format. Need 3 or 6 hex digit.")

    if not is_web(web):
        raise ValueError("Input is not of web type.")
    return rgb2hex(
        [float(int(v)) for v in COLOR_NAME_TO_RGB[web]], force_long
    )  # convert dec to hex


def hsl2hex(hsl: Sequence[int | float]) -> str:
    if not is_hsl(hsl):
        raise ValueError("Input is not of hsl type.")
    return rgb2hex(hsl2rgb(hsl))


def hex2hsl(hex: str) -> tuple[float, float, float]:
    if not (is_long_hex(hex) or is_short_hex(hex)):
        raise ValueError("Input is not of hex type.")
    return rgb2hsl(hex2rgb(hex))


def rgb2web(rgb: Sequence[int | float]) -> str:
    if not is_rgb(rgb):
        raise ValueError("Input is not an RGB type.")
    return hex2web(rgb2hex(rgb))


def web2rgb(web: str) -> tuple[float, float, float]:
    if not is_web(web):
        raise ValueError("Input is not of web type.")
    return hex2rgb(web2hex(web))


def web2hsl(web: str) -> tuple[float, float, float]:
    if not is_web(web):
        raise ValueError("Input is not an web type.")
    return rgb2hsl(web2rgb(web))


def hsl2web(hsl: Sequence[int | float]) -> str:
    if not is_hsl(hsl):
        raise ValueError("Input is not an HSL type.")
    return rgb2web(hsl2rgb(hsl))



from unittest.mock import MagicMock, patch

import pytest

from colourings.colour import (
    HEX,
    HSL,
    RGB,
    Color,
    Colour,
    HSL_equivalence,
    RGB_color_picker,
    color_scale,
    colour_scale,
    identify_color,
    make_color_factory,
)


@patch("tkinter.Tk")
def test_preview(mock_tk):
    c = Colour("red")
    x, y = 300, 300
    mock_root = MagicMock()
    mock_tk.return_value = mock_root
    c.preview(x, y)
    mock_tk.assert_called_once()
    mock_root.geometry.assert_called_once_with(f"{x}x{y}")
    mock_root.config.assert_called_once_with(background=c.hex_l)
    mock_root.title.assert_called_once_with(f"{str(c)} preview")
    mock_root.mainloop.assert_called_once()


def test_preview_invalid_size_x():
    c = Colour("red")
    with pytest.raises(TypeError, match="`size_x` must be of integer or float type"):
        c.preview("invalid", 300)  # type: ignore


def test_preview_invalid_size_y():
    c = Colour("red")
    with pytest.raises(TypeError, match="`size_y` must be of integer or float type"):
        c.preview(300, "invalid")  # type: ignore


@patch("warnings.warn")
def test_preview_alpha_warning(mock_warn):
    c = Colour("red", alpha=0.5)
    with patch("tkinter.Tk"):
        c.preview(300, 300)
    mock_warn.assert_called_once_with(
        f"Alpha set to {c.alpha}, but is not displayed in the window.",
        stacklevel=2,
    )


def test_bad_colour_scale():
    with pytest.raises(ValueError):
        colour_scale((Color("white"),), 2)


def test_colour_scale_with_exact_inputs():
    assert colour_scale((Color("white"), Color("black")), 2) == [
        Color("white"),
        Color("black"),
    ]
    assert colour_scale((Color("blue"), Color("black")), 2) == [
        Color("blue"),
        Color("black"),
    ]
    assert colour_scale((Color("blue"), Color("black"), Color("blue")), 3) == [
        Color("blue"),
        Color("black"),
        Color("blue"),
    ]
    assert colour_scale(
        (Color("blue"), Color("black"), Color("blue"), Color("orange")), 4
    ) == [
        Color("blue"),
        Color("black"),
        Color("blue"),
        Color("orange"),
    ]
    assert colour_scale(
        (Color("blue"), Color("black"), Color("blue"), Color("orange"), Color("green")),
        5,
    ) == [Color("blue"), Color("black"), Color("blue"), Color("orange"), Color("green")]


def test_colour_scale_with_fewer_inputs():
    with pytest.raises(
        ValueError,
        match="Number of steps must be greater than or equal to the number of colors.",
    ):
        colour_scale((Color("white"), Color("black")), 1)
    with pytest.raises(
        ValueError,
        match="Number of steps must be greater than or equal to the number of colors.",
    ):
        colour_scale((Color("blue"), Color("black")), 1)
    with pytest.raises(
        ValueError,
        match="Number of steps must be greater than or equal to the number of colors.",
    ):
        colour_scale((Color("blue"), Color("black"), Color("blue")), 2)


def test_color_scale_with_fewer_inputs():
    with pytest.raises(
        ValueError,
        match="Number of steps must be greater than or equal to the number of colors.",
    ):
        color_scale((Color("white"), Color("black")), 1)
    with pytest.raises(
        ValueError,
        match="Number of steps must be greater than or equal to the number of colors.",
    ):
        color_scale((Color("blue"), Color("black")), 1)
    with pytest.raises(
        ValueError,
        match="Number of steps must be greater than or equal to the number of colors.",
    ):
        color_scale((Color("blue"), Color("black"), Color("blue")), 2)


def test_bad_color_change_HSL():
    c = Color("red")
    with pytest.raises(TypeError, match="Value is not a valid HSL"):
        c.lightness = 2
    with pytest.raises(TypeError, match="Value is not a valid HSL"):
        c.lightness = -0.5
    with pytest.raises(TypeError, match="Value is not a valid HSL"):
        c.saturation = 2
    with pytest.raises(TypeError, match="Value is not a valid HSL"):
        c.saturation = -0.5
    with pytest.raises(TypeError, match="Value is not a valid HSL"):
        c.hue = 361
    with pytest.raises(TypeError, match="Value is not a valid HSL"):
        c.hue = -0.5


def test_bad_color_change_alpha():
    c = Color("red")
    with pytest.raises(ValueError, match="Alpha must be between 0 and 1."):
        c.alpha = 2
    with pytest.raises(ValueError, match="Alpha must be between 0 and 1."):
        c.alpha = -0.5


def test_bad_color_change_rgb():
    c = Color("red")
    with pytest.raises(ValueError, match="Input is not an RGB type."):
        c.red = 300
    with pytest.raises(ValueError, match="Input is not an RGB type."):
        c.red = -0.5
    with pytest.raises(ValueError, match="Input is not an RGB type."):
        c.green = 300
    with pytest.raises(ValueError, match="Input is not an RGB type."):
        c.green = -0.5
    with pytest.raises(ValueError, match="Input is not an RGB type."):
        c.blue = 300
    with pytest.raises(ValueError, match="Input is not an RGB type."):
        c.blue = -0.5


def test_bad_color_scale():
    with pytest.raises(ValueError):
        color_scale((Color("white"),), 2)


def test_color_scale_with_exact_inputs():
    assert color_scale((Color("white"), Color("black")), 2) == [
        Color("white"),
        Color("black"),
    ]
    assert color_scale((Color("blue"), Color("black")), 2) == [
        Color("blue"),
        Color("black"),
    ]
    assert color_scale((Color("blue"), Color("black"), Color("blue")), 3) == [
        Color("blue"),
        Color("black"),
        Color("blue"),
    ]
    assert color_scale(
        (Color("blue"), Color("black"), Color("blue"), Color("orange")), 4
    ) == [
        Color("blue"),
        Color("black"),
        Color("blue"),
        Color("orange"),
    ]
    assert color_scale(
        (Color("blue"), Color("black"), Color("blue"), Color("orange"), Color("green")),
        5,
    ) == [Color("blue"), Color("black"), Color("blue"), Color("orange"), Color("green")]


def test_bad_alpha():
    with pytest.raises(ValueError):
        Color(rgb=(1, 1, 1), alpha=-1)
    with pytest.raises(ValueError):
        Color(rgb=(1, 1, 1), alpha=1.1)
    with pytest.raises(ValueError):
        Color(rgba=(1, 1, 1, 1), alpha=0)
    with pytest.raises(ValueError):
        Color(rgba=(1, 1, 1, 1), alpha=1)
    with pytest.raises(ValueError):
        Color(rgbaf=(1, 1, 1, 1), alpha=0)
    with pytest.raises(ValueError):
        Color(hsla=(1, 1, 1, 1), alpha=0)


def test_alpha_entered_twice():
    c = Color(rgba=(1, 1, 1, 255), alpha=1)
    assert c.alpha == 1
    Color(rgbaf=(1, 1, 1, 1), alpha=1)
    assert c.alpha == 1
    Color(hsla=(1, 1, 1, 1), alpha=1)
    assert c.alpha == 1


def test_bad_identify_color():
    with pytest.raises(TypeError, match="Cannot identify color."):
        identify_color("a")
    with pytest.raises(
        TypeError, match="Cannot determine whether color is RGBA or HSLA."
    ):
        identify_color((0, 0, 0, 0))


def test_RGB():
    assert RGB.WHITE == (255.0, 255.0, 255.0)
    assert RGB.BLUE == (0.0, 0.0, 255.0)
    with pytest.raises(AttributeError):
        RGB.DONOTEXISTS  # noqa: B018


def test_HEX():
    assert HEX.WHITE == "#fff"
    assert HEX.BLUE == "#00f"
    with pytest.raises(AttributeError):
        HEX.DONOTEXISTS  # noqa: B018


def test_color_scale_num_sections():
    n = 10
    cs = color_scale(
        (Color("black"), Color("orange"), Color("blue"), Color("white")), n
    )
    assert cs == [
        Color("black"),
        Color("#39221c"),
        Color("#8e4d1c"),
        Color("orange"),
        Color("#ff003c"),
        Color("#e100ff"),
        Color("blue"),
        Color("#bd71e3"),
        Color("#e3c6d9"),
        Color("white"),
    ]
    assert len(cs) == n

    n = 12
    cs = color_scale(
        (Color("black"), Color("orange"), Color("blue"), Color("white")), n
    )
    assert cs == [
        Color("black"),
        Color("#39221c"),
        Color("#8e4d1c"),
        Color("orange"),
        Color("#ff0004"),
        Color("#ff00ac"),
        Color("#a900ff"),
        Color("blue"),
        Color("#9f58e7"),
        Color("#df9fdf"),
        Color("#e7d7df"),
        Color("white"),
    ]
    assert len(cs) == n

    n = 4
    cs = color_scale((Color(hsl=(0, 1, 0.5)), Color(hsl=(360, 1, 0.5))), n, longer=True)
    assert cs == [Color("#f00"), Color("#0f0"), Color("#00f"), Color("#f00")]
    assert len(cs) == n

    n = 4
    cs = color_scale((Color(hsl=(360, 1, 0.5)), Color(hsl=(0, 1, 0.5))), n, longer=True)
    assert len(cs) == n
    assert cs == [Color("#f00"), Color("#00f"), Color("#0f0"), Color("#f00")]

    n = 4
    cs = color_scale((Color(hsl=(0, 1, 0.5)), Color(hsl=(360, 1, 0.5))), n)
    assert len(cs) == n
    assert cs == [
        Color("#f00"),
        Color("#f00"),
        Color("#f00"),
        Color("#f00"),
    ]

    n = 4
    cs = color_scale((Color(hsl=(360, 1, 0.5)), Color(hsl=(0, 1, 0.5))), n)
    assert len(cs) == n
    assert cs == [
        Color("#f00"),
        Color("#f00"),
        Color("#f00"),
        Color("#f00"),
    ]

    n = 4
    cs = color_scale(
        (Color(hsl=(360.0 / 3, 1, 0.5)), Color(hsl=(2 * 360.0 / 3, 1, 0.5))), n
    )
    assert len(cs) == n
    assert cs == [Color("#0f0"), Color("#0fa"), Color("#0af"), Color("#00f")]

    n = 4
    cs = color_scale(
        (Color(hsl=(360.0 / 3, 1, 0.5)), Color(hsl=(2 * 360.0 / 3, 1, 0.5))),
        n,
        longer=True,
    )
    assert len(cs) == n
    assert cs == [Color("#0f0"), Color("#fa0"), Color("#f0a"), Color("#00f")]

    n = 4
    cs = color_scale(
        (Color(hsl=(2 * 360.0 / 3, 1, 0.5)), Color(hsl=(360.0 / 3, 1, 0.5))),
        n,
        longer=True,
    )
    assert len(cs) == n
    assert cs == [Color("#00f"), Color("#f0a"), Color("#fa0"), Color("#0f0")]

    n = 16
    cs = color_scale((Color(hsl=(0, 0, 0)), Color(hsl=(0, 0, 1))), n)
    assert len(cs) == n
    assert cs == [
        Color("#000"),
        Color("#111"),
        Color("#222"),
        Color("#333"),
        Color("#444"),
        Color("#555"),
        Color("#666"),
        Color("#777"),
        Color("#888"),
        Color("#999"),
        Color("#aaa"),
        Color("#bbb"),
        Color("#ccc"),
        Color("#ddd"),
        Color("#eee"),
        Color("#fff"),
    ]


def test_RGB_color_picker():
    assert RGB_color_picker("Something") == RGB_color_picker("Something")
    assert RGB_color_picker("Something") != RGB_color_picker("Something else")
    assert isinstance(RGB_color_picker("Something"), Color)


def test_colour():
    assert Colour("red") == Color("red")


def test_only_one_input():
    with pytest.raises(ValueError):
        Color(color="red", pick_for="foo")


@pytest.mark.xfail(strict=False)
def test_pick_for():
    foo = object()
    bar = object()
    assert Color(pick_for=foo) == Color(pick_for=foo)
    assert Color(pick_for=foo) != Color(pick_for=bar)


def test_cannot_identify():
    with pytest.raises(TypeError):
        Color((0, 0, 0))
    with pytest.raises(TypeError):
        Color((255, 0, 0))


def test_color_str():
    c = Color("red")
    assert str(c) == "red"
    assert repr(c) == "<Color red>"


def test_purple_inputs():
    assert (
        Color("purple")
        == Color("#800080")
        == Color(hsl=(300, 1, 0.25098039215686274))
        == Color(hsla=(300 / 360, 1, 0.25098039215686274, 1.0))
        == Color((300, 1, 0.25098039215686274))
        == Color(Color("purple"))
    )


def test_red_inputs():
    assert (
        Color("red")
        == Color("blue", hue=0)
        == Color("#f00")
        == Color("#ff0000")
        == Color(hsl=(0, 1, 0.5))
        == Color(hsla=(0, 1, 0.5, 1))
        == Color(rgb=(255, 0, 0))
        == Color(rgba=(255, 0, 0, 255))
        == Color(rgbf=(1, 0, 0))
        == Color(rgbaf=(1, 0, 0, 1))
        == Color(Color("red"))
    )


def test_blue_inputs():
    assert (
        Color("blue")
        == Color("#00f")
        == Color("#0000ff")
        == Color(hsl=(240, 1, 0.5))
        == Color(hsla=(240 / 360, 1, 0.5, 1.0))
        == Color(rgb=(0, 0, 255))
        == Color(rgba=(0, 0, 255, 255))
        == Color(rgbf=(0, 0, 1))
        == Color(rgbaf=(0, 0, 1, 1))
        == Color((0, 0, 255))
        == Color(Color("blue"))
    )


def test_no_eq():
    with pytest.raises(NotImplementedError):
        Color("red") == "red"  # noqa: B015


def test_no_attribute():
    c = Color("red")
    with pytest.raises(AttributeError):
        c.does_not_exists  # noqa: B018
    with pytest.raises(AttributeError):
        c.get_does_not_exists  # noqa: B018


def test_web1():
    red = Color("red")
    blue = Color("blue")
    red.web = "blue"
    assert red == blue


def test_rgb():
    blue1 = Color(rgb=(0, 0, 255))
    blue = Color("blue")
    assert blue1 == blue


def test_hex_l():
    blue1 = Color(hex_l="#0000ff")
    blue = Color("blue")
    assert blue1 == blue


def test_hex():
    blue1 = Color(hex="#00f")
    blue = Color("blue")
    assert blue1 == blue


def test_web():
    blue1 = Color(web="blue")
    blue = Color("blue")
    assert blue1 == blue


def test_get_luminance():
    blue = Color("blue")
    assert round(blue.luminance, 4) == 0.3376


def test_color_range_to():
    red = Color("red")
    blue = Color("blue")
    assert list(red.range_to(blue, 5)) == [
        Color("red"),
        Color("#ff007f"),
        Color("magenta"),
        Color("#7f00ff"),
        Color("blue"),
    ]
    black = Color("black")
    white = Color("white")
    assert list(black.range_to(white, 6)) == [
        Color("black"),
        Color("#333"),
        Color("#666"),
        Color("#999"),
        Color("#ccc"),
        Color("white"),
    ]
    lime = Color("lime")
    assert list(red.range_to(lime, 5)) == [
        Color("red"),
        Color("#ff7f00"),
        Color("yellow"),
        Color("chartreuse"),
        Color("lime"),
    ]


def test_HSL_equivalence():
    black_red = Color("red", hue=0, equality=HSL_equivalence)
    black_blue = Color("blue", hue=0, equality=HSL_equivalence)
    assert black_red == black_blue


def test_color_access():
    b = Color("black")
    b.hsl = HSL.BLUE
    assert round(b.hue / 360.0, 4) == 0.6667
    assert b.saturation == 1.0
    assert b.lightness == 0.5
    assert b.red == 0.0
    assert b.blue == 255.0
    assert b.green == 0.0
    assert b.rgb == (0.0, 0.0, 255.0)
    assert b.rgbf == (0.0, 0.0, 1.0)
    assert b.rgba == (0.0, 0.0, 255.0, 255.0)
    assert b.rgbaf == (0.0, 0.0, 1.0, 1.0)
    assert round(b.hsl[0] / 360.0, 4) == 0.6667
    assert b.hsl[1:] == (1.0, 0.5)
    assert b.hex == "#00f"


def test_color_change_values():
    b = Color("black")
    b.hsl = HSL.BLUE
    b.hue = 0.0
    assert b.hex == "#f00"
    b.hue = 2.0 / 3 * 360.0
    assert b.hex == "#00f"
    b.hex = "#f00"
    assert b.hsl == (0.0, 1.0, 0.5)

    b.hex_l = "#123456"
    assert b.hex_l == "#123456"
    assert b.hex == "#123456"

    b.hex_l = "#ff0000"
    assert b.hex_l == "#ff0000"
    assert b.hex == "#f00"


def test_color_properties():
    c = Color("blue")
    c.hue = 0
    assert c == Color("red")

    c.saturation = 0.0
    assert c.hsl == (0, 0.0, 0.5)
    assert c.rgb == (0.5 * 255.0, 0.5 * 255.0, 0.5 * 255.0)

    c.lightness = 0.0
    assert Color("black") == c
    assert c.hex == "#000"

    c.green = 1.0 * 255.0
    c.blue = 1.0 * 255.0
    assert c.hex == "#0ff"
    assert c == Color("cyan")

    c = Color("blue", lightness=0.75)
    assert c.web == "#7f7fff"

    c = Color("red", red=0.5 * 255.0)
    assert c.web == "#7f0000"


def test_color_recursive_init():
    assert Color("red") == Color(Color(Color("red")))


def test_alpha():
    c = Color("red")
    assert c.alpha == 1
    assert c.rgb == (255.0, 0.0, 0.0)
    assert c.rgba == (255.0, 0.0, 0.0, 255.0)
    assert c.rgbf == (1.0, 0.0, 0.0)
    assert c.rgbaf == (1.0, 0.0, 0.0, 1.0)
    assert c.hsl == (0, 1, 0.5)
    assert c.hsla == (0, 1, 0.5, 1.0)
    c.alpha = 0.5
    assert c.alpha == 0.5
    assert c.rgb == (255.0, 0.0, 0.0)
    assert c.rgba == (255.0, 0.0, 0.0, 127.5)
    assert c.rgbf == (1.0, 0.0, 0.0)
    assert c.rgbaf == (1.0, 0.0, 0.0, 0.5)
    assert c.hsl == (0, 1, 0.5)
    assert c.hsla == (0, 1, 0.5, 0.5)
    with pytest.raises(ValueError):
        c.alpha = -0.1
    with pytest.raises(ValueError):
        c.alpha = 1.1


def test_color_equality():
    assert Color("red") != Color("blue")
    assert Color("red") == Color("red")
    assert Color("red") != Color("blue")
    assert Color("red") == Color("red")


def test_color_equality_change():
    def saturation_equality(c1, c2):
        return c1.lightness == c2.lightness

    assert Color("red", equality=saturation_equality) == Color("blue")


def test_color_subclassing():
    class Tint(Color):
        pass

    assert Tint("red").hsl == (0.0, 1.0, 0.5)


def test_color_factory():
    get_color = make_color_factory(
        equality=HSL_equivalence, picker=RGB_color_picker, pick_key=str
    )
    black_red = get_color("red", lightness=0)
    black_blue = get_color("blue", lightness=0)
    assert isinstance(black_red, Color)
    assert black_red != black_blue



from colourings.definitions import linspace


def test_bad_linspace():
    assert linspace(1, 10, 0) == []


def test_no_end_linspace():
    assert linspace(1, 10, 9, endpoint=False) == [float(i) for i in range(1, 10)]


def test_end_linspace():
    assert linspace(1, 10, 10) == [float(i) for i in range(1, 11)]


def test_linspace_one_num():
    assert linspace(1, 10, 1) == [1]



import pytest

from colourings.conversions import (
    hex2hsl,
    hex2rgb,
    hex2web,
    hsl2hex,
    hsl2rgb,
    hsl2web,
    hsla2hsl,
    rgb2hex,
    rgb2hsl,
    rgb2web,
    rgba2hsl,
    rgbaf2hsl,
    rgbf2hsl,
    web2hex,
    web2hsl,
    web2rgb,
)


def test_rgb2web():
    assert rgb2web((0, 0, 0)) == "black"
    assert rgb2web((0.0, 0.0, 0.0)) == "black"


def test_web2rgb():
    assert web2rgb("black") == (0.0, 0.0, 0.0)


def test_web2hsl():
    hsl = web2hsl("grey")
    assert hsl[0] == 0.0
    assert hsl[1] == 0.0
    assert round(hsl[2], 4) == 0.5020
    hsl1 = web2hsl("grey")
    assert hsl1[0] == 0.0
    assert hsl1[1] == 0.0
    assert round(hsl1[2], 4) == 0.5020


def test_hsl2web():
    assert hsl2web((0.0, 0.0, 0.5020)) == "gray"


def test_hex2hsl():
    hsl = hex2hsl("#00ff00")
    assert round(hsl[0] / 360.0, 4) == 0.3333
    assert hsl[1] == 1.0
    assert hsl[2] == 0.5


def test_hsl2hex():
    assert hsl2hex((1.0, 1.0, 1.0)) == "#fff"


def test_hex2web_7_to_4_digits():
    assert hex2web("#112233") == "#123"


def test_web2hex():
    assert web2hex("#123", True) == "#112233"


def test_bad_rgbaf2hsl():
    with pytest.raises(ValueError):
        rgbaf2hsl("a")  # type: ignore
    with pytest.raises(ValueError):
        rgbaf2hsl((2, 0, 0, 0))
    with pytest.raises(ValueError):
        rgbaf2hsl((0, 2, 0, 0))
    with pytest.raises(ValueError):
        rgbaf2hsl((0, 0, 2, 0))
    with pytest.raises(ValueError):
        rgbaf2hsl((0, 0, 0, 2))


def test_bad_rgba2hsl():
    with pytest.raises(ValueError):
        rgba2hsl("a")  # type: ignore
    with pytest.raises(ValueError):
        rgba2hsl((256, 0, 0, 0))
    with pytest.raises(ValueError):
        rgba2hsl((0, 256, 0, 0))
    with pytest.raises(ValueError):
        rgba2hsl((0, 0, 256, 0))
    with pytest.raises(ValueError):
        rgba2hsl((0, 0, 0, 256))


def test_bad_rgbf2hsl():
    with pytest.raises(ValueError):
        rgbf2hsl("a")  # type: ignore
    with pytest.raises(ValueError):
        rgbf2hsl((1.1, 0, 0, 0))
    with pytest.raises(ValueError):
        rgbf2hsl((0, 1.1, 0, 0))
    with pytest.raises(ValueError):
        rgbf2hsl((0, 0, 1.1, 0))
    with pytest.raises(ValueError):
        rgbf2hsl((0, 0, 0, 1.1))


def test_bad_hsla2hsl():
    with pytest.raises(ValueError):
        hsla2hsl("a")  # type: ignore
    with pytest.raises(ValueError):
        hsla2hsl((2, 0, 0, 0))
    with pytest.raises(ValueError):
        hsla2hsl((0, 2, 0, 0))
    with pytest.raises(ValueError):
        hsla2hsl((0, 0, 2, 0))
    with pytest.raises(ValueError):
        hsla2hsl((0, 0, 0, 2))


def test_bad_hex2web():
    with pytest.raises(ValueError):
        hex2web("black")


def test_bad_web2hex():
    with pytest.raises(AttributeError):
        web2hex("#1234")
    with pytest.raises(ValueError):
        web2hex("123")


def test_bad_web2rgb():
    with pytest.raises(ValueError):
        web2rgb("#1234")
    with pytest.raises(ValueError):
        web2rgb("123")


def test_bad_web2hsl():
    with pytest.raises(ValueError):
        web2hsl("#1234")
    with pytest.raises(ValueError):
        web2hsl("123")


def test_bad_hsl2web():
    with pytest.raises(ValueError):
        hsl2web("a")  # type: ignore
    with pytest.raises(ValueError):
        hsl2web((0, 0, 0, 0))
    with pytest.raises(ValueError):
        hsl2web((361, 0, 0))
    with pytest.raises(ValueError):
        hsl2web((0.4, 1.1, 0))
    with pytest.raises(ValueError):
        hsl2web((0.4, 0, 1.1))


def test_bad_hsl2hex():
    with pytest.raises(ValueError):
        hsl2hex("a")  # type: ignore
    with pytest.raises(ValueError):
        hsl2hex((0, 0, 0, 0))
    with pytest.raises(ValueError):
        hsl2hex((361, 0, 0))
    with pytest.raises(ValueError):
        hsl2hex((0.4, 1.1, 0))
    with pytest.raises(ValueError):
        hsl2hex((0.4, 0, 1.1))


def test_bad_hex2hsl():
    with pytest.raises(ValueError):
        hex2hsl("black")
    with pytest.raises(ValueError):
        hex2hsl("#black")


def test_bad_rgb2web():
    with pytest.raises(ValueError):
        rgb2web("a")  # type: ignore
    with pytest.raises(ValueError):
        rgb2web((1, 0, 0, 0))
    with pytest.raises(ValueError):
        rgb2web((-1, 0, 0))


def test_bad_hsl2rgb():
    with pytest.raises(ValueError):
        hsl2rgb((0, 2, 0))
    with pytest.raises(ValueError):
        hsl2rgb((0, 0, 2))
    with pytest.raises(ValueError):
        hsl2rgb((0, 0, -1))
    with pytest.raises(ValueError):
        hsl2rgb((0, -1, 0))


def test_bad_rgb2hex():
    with pytest.raises(ValueError):
        rgb2hex((-1, 0, 0, 0))
    with pytest.raises(ValueError):
        rgb2hex((260, 0, 0))


def test_bad_rgb2hsl():
    with pytest.raises(ValueError):
        rgb2hsl((0, 0, -1))
    with pytest.raises(ValueError):
        rgb2hsl((0, -1, 0))
    with pytest.raises(ValueError):
        rgb2hsl((-1, 0, 0))


def test_bad_hex2rgb():
    with pytest.raises(ValueError):
        hex2rgb("#00ff000")



from __future__ import annotations

import hashlib
import math
import tkinter
import warnings
from collections.abc import Callable, Generator, Sequence
from typing import Any

from .conversions import (
    hex2hsl,
    hex2rgb,
    hex2web,
    hsl2hsla,
    hsl2rgb,
    hsl2rgbf,
    hsla2hsl,
    rgb2hex,
    rgb2hsl,
    rgb2rgba,
    rgb2rgbaf,
    rgba2hsl,
    rgbaf2hsl,
    rgbf2hsl,
    web2hex,
    web2hsl,
)
from .definitions import COLOR_NAME_TO_RGB, linspace
from .identify import (
    is_hsl,
    is_hsla,
    is_long_hex,
    is_rgb,
    is_rgba,
    is_short_hex,
    is_web,
)


class C_HSL:
    def __getattr__(self, value):
        label = value.lower()
        if label in COLOR_NAME_TO_RGB:
            return rgb2hsl(COLOR_NAME_TO_RGB[label])
        raise AttributeError(f"{self.__class__} instance has no attribute {value}")


HSL = C_HSL()


class C_RGB:
    """RGB colors container. Provides a quick color access."""

    def __getattr__(self, value):
        return hsl2rgb(getattr(HSL, value))


class C_HEX:
    """RGB colors container. Provides a quick color access."""

    def __getattr__(self, value):
        return rgb2hex(getattr(RGB, value))


RGB = C_RGB()
HEX = C_HEX()


def color_scale(
    colors: Sequence[Color | Colour], num_steps: int, longer: bool = False
) -> list[Color]:
    """Create a color scale using many colours via linear interpolation of hsl.

    TODO: implement better interpolation technique: https://www.alanzucconi.com/2016/01/06/colour-interpolation/

    Parameters
    ----------
    colors : Sequence[Color  |  Colour]
        Sequence of Color objects
    nb : int
        Total number of steps
    longer : bool, optional
        Long or short path, by default False

    Yields
    ------
    list[Color]
        List of Color objects

    Raises
    ------
    ValueError
        Number of colors specified must be at least two
    """
    # checks
    if len(colors) < 2:
        raise ValueError("At least two colours are required to make a scale.")
    if len(colors) > num_steps:
        raise ValueError(
            "Number of steps must be greater than or equal to the number of colors."
        )

    # linearly interpolate between colours
    num_sections = len(colors) - 1
    num_steps_per_iter = math.floor((num_steps - len(colors)) / num_sections)
    remainder = ((num_steps - len(colors)) / num_sections) % 1
    out = []
    added = 0
    for i in range(num_sections):
        # colour definitions
        h1, s1, l1 = colors[i].hsl
        h2, s2, l2 = colors[i + 1].hsl
        h1 /= 360.0
        h2 /= 360.0
        if longer == (abs(h1 - h2) < 0.5):
            if h1 < h2:
                h1 += 1
            else:
                h2 += 1

        # number of colours
        num_colors = num_steps_per_iter + 2  # add 2 for start and end colours
        if round(remainder * (i + 1) - added, 7) >= 1:
            num_colors += 1
            added += 1

        # interpolate
        hs = [(v * 360) % 360 for v in linspace(h1, h2, num_colors)]
        ss = linspace(s1, s2, num_colors)
        ls = linspace(l1, l2, num_colors)
        add = [Color(hsl=(_h, _s, _l)) for _h, _s, _l in zip(hs, ss, ls, strict=False)]

        # add to output
        if i == 0:
            out.extend(add)
        else:
            out.extend(add[1:])
    return out


colour_scale = color_scale


def hash_or_str(obj) -> str | int:
    try:
        return hash((type(obj).__name__, obj))
    except TypeError:
        ## Adds the type name to make sure two object of different type but
        ## identical string representation get distinguished.
        return type(obj).__name__ + str(obj)


def RGB_color_picker(obj) -> Color:
    """Build a color representation from the string representation of an object.

    This allows to quickly get a color from some data, with the
    additional benefit that the color will be the same as long as the
    (string representation of the) data is the same.
    """

    ## Turn the input into a by 3-dividable string. SHA-384 is good because it
    ## divides into 3 components of the same size, which will be used to
    ## represent the RGB values of the color.
    digest = hashlib.sha384(str(obj).encode("utf-8")).hexdigest()

    ## Split the digest into 3 sub-strings of equivalent size.
    subsize = int(len(digest) / 3)
    splitted_digest = [digest[i * subsize : (i + 1) * subsize] for i in range(3)]

    ## Convert those hexadecimal sub-strings into integer and scale them down
    ## to the 0..1 range.
    max_value = float(int("f" * subsize, 16))
    components = [
        int(d, 16)  ## Make a number from a list with hex digits
        / max_value  ## Scale it down to [0.0, 1.0]
        for d in splitted_digest
    ]

    return Color(rgb2hex(components))  ## Profit!


def RGB_equivalence(c1: Color, c2: Color) -> bool:
    return c1.hex_l == c2.hex_l


def HSL_equivalence(c1: Color, c2: Color) -> bool:
    return c1._hsl == c2._hsl


def identify_color(
    color: str | Sequence[int | float] | Color | Colour,
) -> Callable[[Any], Any]:
    # checks
    if (
        isinstance(color, Sequence)
        and len(color) == 3
        and is_rgb(color)
        and is_hsl(color)
    ):
        raise TypeError("Cannot determine whether color is RGB or HSL.")
    elif (
        isinstance(color, Sequence)
        and len(color) == 4
        and is_rgba(color)
        and is_hsla(color)
    ):
        raise TypeError("Cannot determine whether color is RGBA or HSLA.")
    else:
        pass

    # identify colour
    if isinstance(color, Color | Colour):
        return lambda x: x.hsl
    elif (
        isinstance(color, str)
        and is_long_hex(color)
        or isinstance(color, str)
        and is_short_hex(color)
    ):
        return hex2hsl
    elif isinstance(color, str) and is_web(color):
        return web2hsl
    elif isinstance(color, Sequence) and is_rgb(color):
        return rgb2hsl
    elif isinstance(color, Sequence) and is_hsl(color):
        return lambda x: x
    # elif isinstance(color, Sequence) and is_rgba(color): NOTE: unreachable
    #     return rgba2hsl
    # elif isinstance(color, Sequence) and is_hsla(color): NOTE: unreachable
    #     return hsla2hsl
    else:
        raise TypeError("Cannot identify color.")


class Color:
    """Abstraction of a color object

    Color object keeps information of a color. It can input/output to different
    format (HSL, RGB, HEX, WEB) and their partial representation.
    """

    _hsl: tuple[float, float, float]  # internal representation
    hsl: tuple[float, float, float]
    hex: str
    hex_l: str
    rgb: tuple[float, float, float]
    rgba: tuple[float, float, float, float]
    rgbf: tuple[float, float, float]
    rgbaf: tuple[float, float, float, float]
    hsla: tuple[float, float, float, float]
    hue: float
    saturation: float
    lightness: float
    luminance: float
    red: float
    green: float
    blue: float
    alpha: float
    web: str

    def __init__(  # noqa: C901
        self,
        color: str | Sequence[int | float] | Color | None = None,
        *,
        web: str | None = None,
        hsl: Sequence[int | float] | None = None,
        hsla: Sequence[int | float] | None = None,
        hex: str | None = None,
        hex_l: str | None = None,
        rgb: Sequence[int | float] | None = None,
        rgba: Sequence[int | float] | None = None,
        rgbf: Sequence[int | float] | None = None,
        rgbaf: Sequence[int | float] | None = None,
        alpha: float | None = None,
        pick_for: Any = None,
        picker: Callable[[Any], Color] = RGB_color_picker,
        pick_key: Callable[[Any], str | int] = hash_or_str,
        equality: Callable[[Color, Color], bool] = RGB_equivalence,
        **kwargs,
    ):
        # checks
        if (
            sum(
                v is not None
                for v in (
                    color,
                    web,
                    hsl,
                    hsla,
                    hex,
                    hex_l,
                    rgb,
                    rgba,
                    rgbf,
                    rgbaf,
                    pick_for,
                )
            )
            != 1
        ):
            raise ValueError(
                "Only one of 'color', 'web', 'hsl', 'hsla', 'hex', 'hex_l', 'rgb', 'rgba', 'rgbf', 'rgbaf' or 'pick_for' may be entered."
            )

        # convert to hsl
        if color is not None:
            func = identify_color(color)
            self.hsl = func(color)
        elif web is not None:
            self.hsl = web2hsl(web)
        elif hsl is not None:
            self.hsl = hsl  # type: ignore
        elif hsla is not None:
            if alpha is not None and alpha != hsla[3]:
                raise ValueError(
                    f"Alpha value defined twice and does not have the same value: alpha={alpha} and alpha of hsla={hsla[3]}"
                )
            self.hsl, alpha = hsla2hsl(hsla), hsla[3]
        elif hex is not None:
            self.hsl = hex2hsl(hex)
        elif hex_l is not None:
            self.hsl = hex2hsl(hex_l)
        elif rgb is not None:
            self.hsl = rgb2hsl(rgb)
        elif rgba is not None:
            if alpha is not None and alpha != rgba[3] / 255.0:
                raise ValueError(
                    f"Alpha value defined twice and does not have the same value: alpha={alpha} and alpha of rgba={rgba[3] / 255.0}"
                )
            self.hsl, alpha = rgba2hsl(rgba), rgba[3] / 255.0
        elif rgbf is not None:
            self.hsl = rgbf2hsl(rgbf)
        elif rgbaf is not None:
            if alpha is not None and alpha != rgbaf[3]:
                raise ValueError(
                    f"Alpha value defined twice and does not have the same value: alpha={alpha} and alpha of rgbaf={rgbaf[3]}"
                )
            self.hsl, alpha = rgbaf2hsl(rgbaf), rgbaf[3]
        elif pick_for is not None:
            self.hsl = web2hsl(picker(pick_key(pick_for)).web)
        # elif isinstance(color, Color):
        #     self.web = web2hsl(color.web)
        else:
            raise ValueError("Input not recognised")

        # set attributes
        self.equality = equality
        self.alpha = alpha if alpha is not None else 1.0
        for k, v in kwargs.items():
            setattr(self, k, v)

    def __getattr__(self, label: str):
        if label.startswith("get_"):
            raise AttributeError(f"'{label}' not found")
        try:
            return getattr(self, "get_" + label)()
        except AttributeError as e:
            raise AttributeError(f"'{label}' not found") from e

    def __setattr__(self, label, value):
        if label not in ["_alpha", "_hsl", "equality"]:
            fc = getattr(self, "set_" + label)
            fc(value)
        else:
            self.__dict__[label] = value

    def get_hsl(self) -> tuple[float, float, float]:
        return self._hsl

    def get_hex(self) -> str:
        return rgb2hex(self.rgb)

    def get_hex_l(self) -> str:
        return rgb2hex(self.rgb, force_long=True)

    def get_rgb(self) -> tuple[float, float, float]:
        return hsl2rgb(self.hsl)

    def get_rgbf(self) -> tuple[float, float, float]:
        return hsl2rgbf(self.hsl)

    def get_rgba(self) -> tuple[float, float, float, float]:
        return rgb2rgba(hsl2rgb(self.hsl), self._alpha)

    def get_rgbaf(self) -> tuple[float, float, float, float]:
        return rgb2rgbaf(hsl2rgb(self.hsl), self._alpha)

    def get_hsla(self) -> tuple[float, float, float, float]:
        return hsl2hsla(self.hsl, self._alpha)

    def get_hue(self) -> float:
        return self.hsl[0]

    def get_saturation(self) -> float:
        return self.hsl[1]

    def get_lightness(self) -> float:
        return self.hsl[2]

    def get_luminance(self) -> float:
        r, g, b = self.get_rgbf()
        return math.sqrt(0.299 * r**2 + 0.587 * g**2 + 0.114 * b**2)

    def get_red(self) -> float:
        return self.rgb[0]

    def get_green(self) -> float:
        return self.rgb[1]

    def get_blue(self) -> float:
        return self.rgb[2]

    def get_alpha(self) -> float:
        return self._alpha

    def get_web(self) -> str:
        return hex2web(self.hex)

    def set_hsl(self, value: Sequence[float]) -> None:
        if not is_hsl(value):
            raise TypeError("Value is not a valid HSL")
        self._hsl = tuple(value)  # type: ignore

    def set_rgb(self, value: Sequence[float]) -> None:
        self.hsl = rgb2hsl(value)

    def set_hue(self, value: float) -> None:
        self.hsl = (value, self.hsl[1], self.hsl[2])

    def set_saturation(self, value: float) -> None:
        self.hsl = (self.hsl[0], value, self.hsl[2])

    def set_lightness(self, value: float) -> None:
        self.hsl = (self.hsl[0], self.hsl[1], value)

    def set_red(self, value: float) -> None:
        self.rgb = (value, self.rgb[1], self.rgb[2])

    def set_green(self, value: float) -> None:
        self.rgb = (self.rgb[0], value, self.rgb[2])

    def set_blue(self, value: float) -> None:
        self.rgb = (self.rgb[0], self.rgb[1], value)

    def set_alpha(self, value: float) -> None:
        if not 0 <= value <= 1:
            raise ValueError("Alpha must be between 0 and 1.")
        self._alpha = value

    def set_hex(self, value: str) -> None:
        self.rgb = hex2rgb(value)

    def set_hex_l(self, value: str) -> None:
        self.set_hex(value)

    def set_web(self, value: str) -> None:
        self.hex = web2hex(value)

    def range_to(
        self, value: str | Sequence[int | float] | Color, steps, longer=False
    ) -> Generator[Color, None, None]:
        """range of color generation"""
        yield from color_scale((self, Color(value)), steps, longer=longer)

    def preview(self, size_x: int | float = 200, size_y: int | float = 200) -> None:
        if not isinstance(size_x, int | float):
            raise TypeError("`size_x` must be of integer or float type")
        if not isinstance(size_y, int | float):
            raise TypeError("`size_y` must be of integer or float type")
        if self._alpha != 1:
            warnings.warn(
                f"Alpha set to {self._alpha}, but is not displayed in the window.",
                stacklevel=2,
            )
        root = tkinter.Tk()
        root.geometry(f"{size_x}x{size_y}")
        root.config(background=self.get_hex_l())
        root.title(f"{str(self)} preview")
        root.mainloop()

    def __str__(self) -> str:
        return f"{self.web}"

    def __repr__(self) -> str:
        return f"<Color {self.web}>"

    def __eq__(self, other) -> bool:
        if isinstance(other, Color):
            return self.equality(self, other)
        raise NotImplementedError("Other object must be of type `Color` or `Colour`")


class Colour(Color): ...


def make_color_factory(**kwargs_defaults):
    def ColorFactory(*args, **kwargs):
        new_kwargs = kwargs_defaults.copy()
        new_kwargs.update(kwargs)
        return Color(*args, **new_kwargs)

    return ColorFactory
